#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <gsl/gsl_wavelet.h>
#include <gsl/gsl_wavelet2d.h>
#include "mdldenoise.h"
#include "boxmuller.h"

void usage(char *cmd)
{
  fprintf(stderr, "usage: %s <file> [<noise std.dev.>]\n", cmd);
  exit(-1);
}

Signal *read_sig(char *fname)
{
  Signal *y;
  FILE *f;
  int c;
  double val;

  f = fopen(fname, "r");

  y = alloc_signal(1, 8);
  
  c = 0;
  while (!feof(f))
  {
    if (fscanf(f, "%lf", &val) != 1)
      break;
    if (c >= y->len)
      y = realloc_signal(y, y->len*2);
    y->a[c] = val;
    c++;
  }

  if (c < y->len)
    fprintf(stderr, "signal of length %d. padding with zeros to %d.\n",
	    c, y->len);
  for (;c < y->len; c++)
    y->a[c] = 0.0;

  fclose(f);

  return y;
}

void write_sig(FILE *f, Signal *y)
{
  int c; for (c = 0; c < y->len; c++) fprintf(f, "%lf\n", y->a[c]);
}

Signal *read_signal(char *fname)
{
  if (is_png_fname(fname))
    return read_png(fname);
  else
    return read_sig(fname);
}

void write_signal(char *fname, Signal *y)
{
  FILE *f;
  
  if (y->dim == 1)
  {
    f = fopen(fname, "w");
    write_sig(f, y);
    fclose(f);
  }
  else
    write_png(fname, y);
}

Signal *copy_signal(Signal *src)
{
  Signal *dest;
  int row;

  dest = (Signal *) malloc(sizeof(Signal));
  dest->dim = src->dim;
  dest->len = src->len;
  dest->n = src->n;
  dest->a = (double *) malloc(sizeof(double)*dest->n);
  if (dest->dim == 1)
  {
    dest->ap = (double **) malloc(sizeof(double *));
    dest->ap[0] = dest->a;
  }
  else
  {
    dest->ap = (double **) malloc(sizeof(double *)*dest->len);
    for (row = 0; row < dest->len; row++)
      dest->ap[row] = dest->a+row*dest->len;
  }
  memcpy(dest->a, src->a, sizeof(double)*dest->n);

  return dest;
}

Signal *distort_signal(Signal *y, double sigma)
{
  int row, col;

  for (row = 0; row < (y->len-1)*(y->dim-1)+1; row++)
    for (col = 0; col < y->len; col++)
      y->ap[row][col] += boxmuller()*sigma;

  return y;
}

double rmse(Signal *x, Signal *y)
{
  int row, col;
  double ss = 0.0;

  for (row = 0; row < (y->len-1)*(y->dim-1)+1; row++)
    for (col = 0; col < y->len; col++)
      ss += square(x->ap[row][col]-y->ap[row][col]);

  return sqrt(ss/x->n);
}

double log10(double x)
{
  return log(x)/2.302585092994046;
}

double psnr(Signal *x, Signal *y)
{
  int row, col;
  double ss = 0.0, min, max;

  for (row = 0; row < (y->len-1)*(y->dim-1)+1; row++)
    for (col = 0; col < y->len; col++)
    {
      ss += square(x->ap[row][col]-y->ap[row][col]);
      if (y->dim == 1)
      {
	if ((col == 0 && row == 0) || x->ap[row][col] < min)
	  min = x->ap[row][col];
	if ((col == 0 && row == 0) || x->ap[row][col] > max)
	  max = x->ap[row][col];
      }
    }

  if (y->dim == 2)
  {
    min = 0.0;
    max = 255.0;
  }

  return 10.0*log10((max-min)*(max-min)/(ss/x->n));
}

void free_signal(Signal *y)
{
  free(y->ap);
  free(y->a);
  free(y);
}

void wt(Signal *y, int dir)
{
  static gsl_wavelet *w = NULL;
  static gsl_wavelet_workspace *ww = NULL;

  if (!y)
  {
    if (w) gsl_wavelet_free(w);
    if (ww) gsl_wavelet_workspace_free(ww);
    return;
  }

  //if (!w) w = gsl_wavelet_alloc(gsl_wavelet_haar, 2);
  if (!w) w = gsl_wavelet_alloc(gsl_wavelet_daubechies, 6);
  if (!ww) ww = gsl_wavelet_workspace_alloc(y->len);

  if (y->dim == 1)
    gsl_wavelet_transform(w, y->a, 1, y->len, dir, ww);
  else
    gsl_wavelet2d_nstransform(w, y->a, y->len, y->len, y->len, dir, ww);
}

int log2dyad(int x)
{
  int y;
  for (y = -1; x > 0; x >>= 1) y++;
  return y;
}

double square(double x)
{
  return x*x;
}

double lchoose(int n, int k)
{
  if (k < 0 || k > n)
    return -1.0/0.0;

  if (k == 0 || k == n)
    return 0.0;

  //return lgamma(n+1)-lgamma(k+1)-lgamma(n-k+1);

  // Stirling's approximation
  return (n+.5)*log(n)-(k+.5)*log(k)-(n-k+.5)*log(n-k)-.399089934179;
  // .399... ~= .5*log(2*pi)
}

double lchoose2(int n, int k, int l)
{
  if (k < 0 || l < 0 || k+l > n)
    return -1.0/0.0;

  if (MAX(k,l) == 0 || k+l == n)
    return 0.0;

  return lgamma(n+1)-lgamma(k+1)-lgamma(l+1)-lgamma(n-k-l+1);
}

int comp_abs(const void *p1, const void *p2)
{
  double v1, v2;
  v1 = fabs(*((double *)p1));
  v2 = fabs(*((double *)p2));
  if (v1 < v2) return -1;
  else if (v1 > v2) return 1;
  else return 0;
}

int comp_Ival(const void *p1, const void *p2)
{
  double v1, v2;
  v1 = fabs(((Ival *)p1)->val);
  v2 = fabs(((Ival *)p2)->val);
  if (v1 < v2) return 1;
  else if (v1 > v2) return -1;
  else return 0;
}

double estimate_mad(Signal *y)
{
  double *c;
  double sigmahat;
  int row, col, i;

  c = (double *) malloc(sizeof *c*(y->n-1)/(2*y->dim-1)+1);

  i = 0;
  for (row = (y->dim-1)*(y->len/2); 
       row < (y->dim-1)*(y->len-1)+1; row++)
      for (col = y->len/2; col < y->len; col++)
        c[i++] = y->ap[row][col];

  qsort(c, y->n/(2*y->dim), sizeof *c, comp_abs);

  sigmahat = (fabs(c[y->n/(4*y->dim)-1])+fabs(c[y->n/(4*y->dim)]))/2/.6745;

  return sigmahat;
}

double soft_threshold(double val, double thres)
{
  double sign;

  if (val < 0.0)
    sign = -1.0;
  else
    sign = 1.0;

  if (isinf(thres))
    return 0.0;

  if (fabs(val) > thres)
    return val - sign*thres; // soft
  else
    return 0.0;
}

double hard_threshold(double val, double thres)
{
  if (isinf(thres))
    return 0.0;

  if (fabs(val) > thres)
    return val;
  else
    return 0.0;
}

// returns number of non-zero coefficient after thresholding
int threshold_subband(Range *range, Signal *y, double thres,
		      double(*thresf)(double,double))
{
  int row, col, c;

  c = 0;
  for_each_coeff(row, col, *range)
  {
    y->ap[row][col] = (double)(*thresf)(y->ap[row][col], thres);
    if (y->ap[row][col] != 0.0)
      c++;
  }
  return c;
}

int select_subband(Range *range, Signal *y, int level, int band)
{
  Range tmp;

  if (!range)
    range = &tmp;

  if (level < 0) // level -1 = all coefficients
  {
    range->min_row = range->min_col = 0;
    range->max_row = y->n/y->len;
    range->max_col = y->len;
  }
  else if (level == 0) // level 0 = total average (scaled)
  {
    range->min_row = range->min_col = 0;
    range->max_row = range->max_col = 1;
  }
  else
  {
    if (y->dim == 1)
    {
      range->min_row = 0;
      range->max_row = 1;
      range->min_col = 1 << (level-1);
      range->max_col = 1 << level;
    }
    else
    {
      switch (band) {
      case HH: // diagonal (bottom-right)
	range->min_row = range->min_col = 1 << (level-1);
	range->max_row = range->max_col = 1 << level;
	break;
      case HL: // horizontal (top-right)
	range->min_row = 0;
	range->max_row = range->min_col = 1 << (level-1);
	range->max_col = 1 << level;
	break;
      case LH: // vertical (bottom-left)
      default:
	range->min_col = 0;
	range->min_row = range->max_col = 1 << (level-1);
	range->max_row = 1 << level;
      }
    }
  }
  range->n = (range->max_row-range->min_row)*(range->max_col-range->min_col);

  return range->n;
}

// sorts coefficients within 'range' in descending order according to
// absolute value. returns sum of squares
double sort_range(Ival *c, Signal *y, Range *range)
{
  int row, col, i = 0;
  double ss = 0.0;
  for_each_coeff(row, col, *range)
  {
    c[i].row = row;
    c[i].col = col;
    c[i].val = y->ap[row][col];
    ss += square(c[i].val);
    i++;
  }
  qsort(c, range->n, sizeof *c, comp_Ival);
  return ss;
}

void simple_denoise(Signal *y, int method)
{
  Range range;
  int i, k, minlevel, maxlevel, level, band, row, col;
  double mad, ss, thres, gam, out, in, val, sure, best, crit;
  Ival *c;

  if (method == VISU || method == SURE || method == BAYES) // use MAD
    mad = estimate_mad(y);
  else
    mad = 0.0;

  if (method == VISU || method == RNML || method == INML) // non-subband-wise
    minlevel = maxlevel = -1;
  else
  {
    minlevel = BOTTOM_LEVEL;
    maxlevel = log2dyad(y->len);
  }

  for_each_band(level, band, y->dim, minlevel, maxlevel)
  {
    select_subband(&range, y, level, band);
    
    if (method == VISU)
    {
      /*****************************************************************
       *              VisuShrink (Donoho-Johnstone, 1994)              *
       *****************************************************************/
      thres = mad*sqrt(2*log(range.n));
      threshold_subband(&range, y, thres, hard_threshold);
    }
    else if (method == SURE) 
    {
      /*****************************************************************
       *           SureShrink (hybrid) (Donoho-Johnstone, 1995)        *
       *****************************************************************/
      ss = 0.0;
      for_each_coeff(row, col, range)
	ss += square(y->ap[row][col]/mad)-1;
      ss /= range.n;
      gam = log2dyad(range.n); gam *= sqrt(gam/range.n);
      if (ss <= gam)
	thres = mad*sqrt(2*log(range.n));
      else
      {
	c = malloc(sizeof *c*range.n);
	out = sort_range(c, y, &range)/mad;
	in = 0.0;
	for (i = 0; i <= range.n; i++)
	{
	  if (i == 0)
	    val = 0.0;
	  else
	  {
	    val = c[range.n-i].val/mad;
	    in += val*val;
	  }
	  if (i == 0 || i == range.n || 
	      fabs(c[range.n-i].val) != fabs(c[range.n-i+1].val))
	  {
	    sure = range.n - 2*i + (out - in + val*val*i);   // Eq. (11)
	    if (i == 0 || !finite(best) || (finite(sure) && sure < best))
	    {
	      best = sure;
	      thres = fabs(val)*mad;
	    }
	  }
	}
	free(c);
      }
      threshold_subband(&range, y, thres, soft_threshold);
    }
    else if (method == BAYES) 
    {
      /*****************************************************************
       *               BayesShrink (Chang-Yu-Vetterli, 2000)           *
       *****************************************************************/

      ss = 0.0;
      for_each_coeff(row, col, range)
	ss += square(y->ap[row][col]);
      ss /= range.n;
      thres = mad*mad/sqrt(MAX(ss-mad*mad,0.0));
      threshold_subband(&range, y, thres, soft_threshold);
    }
    else if (method == RNML || method == INML)
    {
      /*****************************************************************
       *     MDL denoising (Rissanen, 2000) (Roos et al., 200?)        *
       *****************************************************************/
      
      c = malloc(sizeof *c*range.n);
      out = sort_range(c, y, &range);

      for (i = 1; i < .95*range.n; i++)
      {
	in += square(c[i-1].val);
	out -= square(c[i-1].val);
	if (method == RNML)
	  // Rissanen's formula (Rissanen, 2000)
	  crit = (range.n-i)*log(out/(range.n-i)) + i*log(in/i)
	    + log(i)+log(range.n-i);
	else // (method == INML)
	{
	  // code-length for model index (Roos et al. 20??)

	  crit = (range.n-i)*log(out) + i*log(in)
            - 3*i*log(i) - 3*(range.n-i)*log(range.n-i)
	    - log(i)-log(range.n-i);
	}

	if (i == 1 || !finite(best) || (finite(crit) && crit < best))
	{
	  best = crit;
	  k = i;
	  thres = fabs(c[i].val);
	}
      }

      threshold_subband(&range, y, thres, hard_threshold);
    }
  }
}

double codelength_flip(Signal *y, Ival ***c, double ***Sk, int **k,
		       int minlevel, int maxlevel,
		       int level, int band, int i, int into)
{
  int lev, ban, k0, add1;
  double cl, in0, in;
  Range range;

  k0 = 0; in0 = 0.0; // noise coefficients (level L=0)
  cl = 0.0; 

  for_each_band(lev, ban, y->dim, minlevel, maxlevel)
  {
    select_subband(&range, y, lev, ban);

    in = Sk[lev][ban][k[lev][ban]];
    add1 = 0;
    if (into != 0 && lev == level && ban == band)
    {
      if (i < k[lev][ban] && into == -1) // i is included -> take it out
	add1 = -1;
      else if (i >= k[lev][ban] && into == 1)
        // excluded -> put it in
	add1 = 1;
      in += add1*square(c[lev][ban][i].val);
    }

    if (k[lev][ban]+add1 > 0)
    {
      cl += (k[lev][ban]+add1)*(log(in/(k[lev][ban]+add1)))
	+ log(k[lev][ban]+add1);
      cl += 2*lchoose(range.n, k[lev][ban]+add1);
    }

    // noise

    k0 += range.n-(k[lev][ban]+add1);
    in0 += Sk[lev][ban][range.n]-in;
  }

  if (k0>0)
    cl += k0*(log(in0/k0)) + log(k0);

  return cl;
}

double codelength(Signal *y, Ival ***c, double ***Sk, int **k, 
		  int minlevel, int maxlevel)
{
  return codelength_flip(y, c, Sk, k, minlevel, maxlevel, -1, -1, -1, 0);
}

void iterative_denoise(Signal *y, int method)
{
  int level, band, minlevel, maxlevel, i, changed, **k, bk,
    oldbk, maxk, print_weights = 0;
  double ***Sk, cl, bcl, r1, r2, w;
  Range range;
  Ival ***c;
  FILE *f;

  // initializations

  minlevel = BOTTOM_LEVEL;
  maxlevel = log2dyad(y->len);

  c = malloc(sizeof *c*(maxlevel+1));        // coefficients [level][band][i]
  Sk = malloc(sizeof *Sk*(maxlevel+1));      // S_k (sum c[0...k-1])
  k = malloc(sizeof *k*(maxlevel+1));        // current k
  
  for (level = minlevel; level <= maxlevel; level++)
  {
    c[level] = malloc(sizeof **c*4);
    Sk[level] = malloc(sizeof **Sk*4);
    k[level] = malloc(sizeof **k*4);
  }

  for_each_band(level, band, y->dim, minlevel, maxlevel)
  {
    select_subband(&range, y, level, band);

    c[level][band] = malloc(sizeof ***c*range.n);
    Sk[level][band] = malloc(sizeof ***Sk*(range.n+1));

    sort_range(c[level][band], y, &range);
    Sk[level][band][0] = 0.0;
    for (i = 1; i <= range.n; i++)
      Sk[level][band][i] = Sk[level][band][i-1]+
	square(c[level][band][i-1].val);
    k[level][band] = range.n;
  }

  // optimize k for all bands

  changed = 1;
  while (changed)
  {
    changed = 0;
    for_each_band_reverse(level, band, y->dim, minlevel, maxlevel)
    {
      select_subband(&range, y, level, band);
      oldbk = k[level][band];
      maxk = .95*range.n;
      for (k[level][band] = 0; k[level][band] <= maxk; k[level][band]++)
      {
	cl = codelength(y, c, Sk, k, minlevel, maxlevel);
	if (k[level][band] == 0 || !finite(bcl) || (finite(cl) && cl < bcl))
	{
	  bk = k[level][band];
	  bcl = cl;
	}
      }
      k[level][band] = bk;
      if (0 && method == SNML)
	fprintf(stderr, "%d %d %d->%d/%d/%d %s %.1f\n", 
		level, band, oldbk, bk, maxk, range.n, bk!=oldbk?"***":"",
		bcl);
      if (bk != oldbk)
	changed = 1;
    }
  }

  // threshold

  if (method == MNML && print_weights)
    f = fopen("weights.dat", "w");

  for_each_band(level, band, y->dim, minlevel, maxlevel)
  {
    select_subband(&range, y, level, band);
    for (i = 0; i < range.n; i++)
    {
      if (method == SNML)
      {
	// hard threshold
	if (i >= k[level][band]) 
	  w = 0.0;
	else
	  w = 1.0;
      }
      else
      {
	r1 = codelength_flip(y, c, Sk, k, minlevel, maxlevel, 
			     level, band, i, 1);
	r2 = codelength_flip(y, c, Sk, k, minlevel, maxlevel, 
			     level, band, i, -1);
	r1 /= 2; r2 /= 2;
	bk = k[level][band];
	k[level][band] = i+1;
	if (print_weights)
	  fprintf(f, "%d %d %d %f %f ", level, band, i,
		  codelength(y, c, Sk, k, minlevel, maxlevel), r1-r2);
	k[level][band] = bk;

	if (r1-r2 < -10 || !finite(r2))
	  w = 1.0;
	else if (r1-r2 > 10 || !finite(r1))
	  w = 0.0;
	else
	{
	  w = exp(-(r1-r2));
	  w = w/(1+w);
	}
	if (print_weights)
	  fprintf(f, "%f\n", w);
      }

      y->ap[c[level][band][i].row][c[level][band][i].col] = 
	w*c[level][band][i].val;
    }
  }

  if (method == MNML && print_weights)
    fclose(f);
  
  // clean up

  for_each_band(level, band, y->dim, minlevel, maxlevel)
  {
    free(c[level][band]);
    free(Sk[level][band]);
  }
  for (level = minlevel; level <= maxlevel; level++)
  {
    free(c[level]);
    free(Sk[level]);
    free(k[level]);
  }
  free(c);
  free(Sk);
  free(k);
}

void denoise(Signal *y, int method)
{
  if (method == VISU || method == SURE || method == BAYES ||
      method == RNML || method == INML)
    simple_denoise(y, method);
  else 
    iterative_denoise(y, method);
}

char *get_fname(char *body, char *ext)
{
  static char fname[256];
  char *p;

  strcpy(fname, body);
  p = strrchr(fname, '.');
  if (!p) p = fname+strlen(fname);

  *p = '_';
  strcpy(p+1, ext);
  strcpy(p+1+strlen(ext), body+(p-fname));

  return fname;
}

int main(int argc, char *argv[])
{
  Signal *orig, *noisy, *tmp;
  int nofiles = 0;
  double sigma = 0.0;

  if (argc < 2) usage(argv[0]);
  if (argc > 2) sigma = atof(argv[2]);
  if (argc > 3 && strcmp(argv[3], "--nofiles") == 0) nofiles = 1;

  srand48(4);
  _boxmuller_setrandomseed();

  orig = read_signal(argv[1]);
  noisy = distort_signal(copy_signal(orig), sigma); 
  if (!nofiles) write_signal(get_fname(argv[1], "noisy"), noisy);
  printf ("none %f %f\n", rmse(orig, noisy), psnr(orig, noisy));
  wt(noisy, forward);

  tmp = copy_signal(noisy); denoise(tmp, VISU); wt(tmp, backward);
  printf("visu %f %f\n", rmse(orig, tmp), psnr(orig, tmp));
  if (!nofiles) write_signal(get_fname(argv[1], "visu"), tmp); 
  free_signal(tmp);

  tmp = copy_signal(noisy); denoise(tmp, SURE); wt(tmp, backward);
  printf("sure %f %f\n", rmse(orig, tmp), psnr(orig, tmp));
  if (!nofiles) write_signal(get_fname(argv[1], "sure"), tmp);
  free_signal(tmp);

  tmp = copy_signal(noisy); denoise(tmp, BAYES); wt(tmp, backward);
  printf("bayes %f %f\n", rmse(orig, tmp), psnr(orig, tmp));
  if (!nofiles) write_signal(get_fname(argv[1], "bayes"), tmp);
  free_signal(tmp);

  tmp = copy_signal(noisy); denoise(tmp, RNML); wt(tmp, backward);
  printf("rnml %f %f\n", rmse(orig, tmp), psnr(orig, tmp)); 
  if (!nofiles) write_signal(get_fname(argv[1], "rnml"), tmp);
  free_signal(tmp);

  tmp = copy_signal(noisy); denoise(tmp, INML); wt(tmp, backward);
  printf("inml %f %f\n", rmse(orig, tmp), psnr(orig, tmp));
  if (!nofiles) write_signal(get_fname(argv[1], "inml"), tmp);
  free_signal(tmp); 

  tmp = copy_signal(noisy); denoise(tmp, SNML); wt(tmp, backward);
  printf("snml %f %f\n", rmse(orig, tmp), psnr(orig, tmp));
  if (!nofiles) write_signal(get_fname(argv[1], "snml"), tmp);
  free_signal(tmp);

  tmp = copy_signal(noisy); denoise(tmp, MNML); wt(tmp, backward);
  printf("mnml %f %f\n", rmse(orig, tmp), psnr(orig, tmp));
  if (!nofiles) write_signal(get_fname(argv[1], "mnml"), tmp);
  free_signal(tmp);

  free_signal(noisy);
  free_signal(orig);
  wt(NULL, 0);

  return 0;
}
